#include "include.h"
//global functions are good.
PFNWGLGETEXTENSIONSSTRINGARBPROC	wglGetExtensionsStringARB;
PFNWGLCHOOSEPIXELFORMATARBPROC		wglChoosePixelFormatARB;
PFNGLACTIVETEXTUREARBPROC		glActiveTexture;
PFNWGLGETPIXELFORMATATTRIBFVARBPROC	wglGetPixelFormatAttribfvARB;
PFNWGLGETPIXELFORMATATTRIBIVARBPROC	wglGetPixelFormatAttribivARB;
PFNGLGENPROGRAMSARBPROC			glGenProgramsARB;
PFNGLBINDPROGRAMARBPROC			glBindProgramARB;
PFNGLPROGRAMSTRINGARBPROC		glProgramStringARB;

//GLSL
PFNGLCREATEPROGRAMPROC			glCreateProgram;
PFNGLLINKPROGRAMPROC			glLinkProgram;
PFNGLGETPROGRAMIVPROC			glGetProgramiv;
PFNGLGETPROGRAMINFOLOGPROC		glGetProgramInfoLog;
PFNGLVALIDATEPROGRAMPROC		glValidateProgram;
PFNGLUSEPROGRAMPROC			glUseProgram;
PFNGLCREATESHADERPROC			glCreateShader;
PFNGLSHADERSOURCEPROC			glShaderSource;
PFNGLCOMPILESHADERPROC			glCompileShader;
PFNGLGETSHADERIVPROC			glGetShaderiv;
PFNGLGETSHADERINFOLOGPROC		glGetShaderInfoLog;
PFNGLATTACHSHADERPROC			glAttachShader;
PFNGLDETACHSHADERPROC			glDetachShader;
PFNGLDELETESHADERPROC			glDeleteShader;
PFNGLGETUNIFORMLOCATIONPROC		glGetUniformLocation;
PFNGLUNIFORM1IPROC			glUniform1i;
PFNGLUNIFORM1FPROC			glUniform1f;
PFNGLUNIFORM2FPROC			glUniform2f;
PFNGLUNIFORM3FPROC			glUniform3f;
PFNGLUNIFORM4FPROC			glUniform4f;


//global definitions are bad
NxPhysicsSDK	*gPhysicsSDK = NULL;
NxScene		*gScene = NULL;
NxSceneDesc	sceneDesc;
GLuint		dispListSphere;
GLuint		dispListCube;
GLuint		texObject[6];
BOOL		bMultiSample = FALSE;

LRESULT CALLBACK mainProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK initProc(HWND, UINT, WPARAM, LPARAM);

BOOL setupNewPixelFormat(HDC);
BOOL setupPixelFormat(HDC);
int windowMain(HINSTANCE, HINSTANCE, PSTR, INT);
int windowInit(HINSTANCE, HINSTANCE, LPSTR, INT);

void initGL(void);
void initNx(void);
void initWGLExt(HDC);
int initExt(void);
void drawScene(HWND, HDC, POINT);
void resize(GLint, GLint);


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, INT iCmdShow)
{
	windowInit(hInstance, hPrevInstance, szCmdLine, iCmdShow);
	windowMain(hInstance, hPrevInstance, szCmdLine, iCmdShow);
	//SetDeviceGammaRamp
	//GetDeviceGammaRamp
	return 0;
}

int windowInit(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR szCmdLine, INT iCmdShow)
{
	HWND		hwnd;
	MSG		msg;
	WNDCLASS	wndclass;
	const CHAR	szAppName[] = TEXT("winInit");

	wndclass.style		= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= initProc;
	wndclass.cbClsExtra	= 0;
	wndclass.cbWndExtra	= 0;
	wndclass.hInstance	= hInstance;
	wndclass.hIcon		= LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor	= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH) GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName	= NULL;
	wndclass.lpszClassName	= szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, "Unable to register window class.", szAppName, MB_ICONERROR);
		return 0;
	}

	hwnd = CreateWindow(	szAppName,			// window class name
				szAppName,			// window caption
				WS_OVERLAPPEDWINDOW,		// window style
				CW_USEDEFAULT,			// initial x position
				CW_USEDEFAULT,			// initial y position
				CW_USEDEFAULT,			// initial x size
				CW_USEDEFAULT,			// initial y size
				NULL,				// parent window handle
				NULL,				// window menu handle
				hInstance,			// program instance handle
				NULL);				// creation parameters

//	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	if (!UnregisterClass(szAppName, hInstance))
	{
		MessageBox(NULL, TEXT("Unable to unregister class."), szAppName, MB_ICONERROR);
		return 1;
	}
	return msg.wParam;
}

LRESULT CALLBACK initProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HDC	hdc;
	static HGLRC	hglrc;

	switch (message)
	{
	case WM_CREATE:
		hdc = GetDC(hwnd);
		setupPixelFormat(hdc);
		hglrc = wglCreateContext(hdc);
		wglMakeCurrent(hdc, hglrc);
		initWGLExt(hdc);
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		return 0;
	case WM_DESTROY:
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(hglrc);
		ReleaseDC(hwnd, hdc);
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}

void initWGLExt(HDC hdc)
{
	const CHAR	*sExt;

	wglGetExtensionsStringARB	= (PFNWGLGETEXTENSIONSSTRINGARBPROC) wglGetProcAddress("wglGetExtensionsStringARB");
	if (wglGetExtensionsStringARB)
	{
		sExt = wglGetExtensionsStringARB(hdc);

		if (strstr("WGL_ARB_pixel_format ", sExt) && strstr("WGL_ARB_multisample ", sExt) )
		{
			bMultiSample = TRUE;
			wglChoosePixelFormatARB		= (PFNWGLCHOOSEPIXELFORMATARBPROC) wglGetProcAddress("wglChoosePixelFormatARB");
			wglGetPixelFormatAttribfvARB	= (PFNWGLGETPIXELFORMATATTRIBFVARBPROC) wglGetProcAddress("wglGetPixelFormatAttribfvARB");
			wglGetPixelFormatAttribivARB	= (PFNWGLGETPIXELFORMATATTRIBIVARBPROC) wglGetProcAddress("wglGetPixelFormatAttribivARB");
		}
	}
}

int windowMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR szCmdLine, INT iCmdShow)
{
	HWND		hwnd;
	HACCEL		hAccel;
	MSG		msg;
	WNDCLASS	wndclass;
	BOOL		running = TRUE;
	char		szAppName[] = TEXT("WinMain");

	wndclass.style		= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndclass.lpfnWndProc	= mainProc;
	wndclass.cbClsExtra	= 0;
	wndclass.cbWndExtra	= 0;
	wndclass.hInstance	= hInstance;
	wndclass.hIcon		= LoadIcon(hInstance, szAppName);
	wndclass.hCursor	= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= NULL;
	wndclass.lpszMenuName	= NULL;
	wndclass.lpszClassName	= szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Unable to register window class."), szAppName, MB_ICONERROR);
		return 0;
	}
	hwnd = CreateWindow(	szAppName,			// window class name
				szAppName,			// window caption
				WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,		// window style
				CW_USEDEFAULT,			// initial x position
				CW_USEDEFAULT,			// initial y position
				CW_USEDEFAULT,			// initial x size
				CW_USEDEFAULT,			// initial y size
				NULL,				// parent window handle
				NULL,				// window menu handle
				hInstance,			// program instance handle
				NULL);				// creation parameters

	hAccel = LoadAccelerators(hInstance, szAppName);
	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (running)
	{
		while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if ( GetMessage(&msg, NULL, 0, 0) > 0 )
			{
				if (!TranslateAccelerator(hwnd, hAccel, &msg))
				{
					TranslateMessage (&msg) ;
					DispatchMessage (&msg) ;
				}
			} else {
				running = FALSE;
			}
		}
		SendMessage(hwnd, WMU_IDLE, 0, 0);
	}

	if (!UnregisterClass(szAppName, hInstance))
	{
		MessageBox(NULL, TEXT("Unable to unregister class."), szAppName, MB_ICONERROR);
		return 1;
	}
	return msg.wParam ;
}

LRESULT CALLBACK mainProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HDC		hdc;
	static HGLRC		hglrc;
	static BOOL		captured = FALSE;
	static POINT		mousePos;

	switch (message)
	{
	case WMU_IDLE:
		drawScene(hwnd, hdc, mousePos);
		return 0;
	case WM_CREATE:
		hdc = GetDC(hwnd);
		if (bMultiSample)
			setupNewPixelFormat(hdc);
		else
			setupPixelFormat(hdc);

		hglrc = wglCreateContext(hdc);
		wglMakeCurrent(hdc, hglrc);
		if (initExt())
		{
			MessageBox(NULL, "OpenGL version 2.0 or higher required.", "Error", MB_OK);
			wglMakeCurrent(NULL, NULL);
			wglDeleteContext(hglrc);
			ReleaseDC(hwnd, hdc);
			PostQuitMessage(0);
			return 0;
		}
		wglUseFontBitmaps(hdc, 0, 256, 1000);
		initNx();
		initGL();
		return 0;

	case WM_LBUTTONDOWN:
		if (!captured)
		{
			SetCapture(hwnd);
			captured = TRUE;
		}
		return 0;

	case WM_LBUTTONUP:
		if (captured)
		{
			ReleaseCapture();
			captured = FALSE;
		}
		return 0;

	case WM_MOUSEMOVE:
		mousePos.x = LOWORD(lParam);
		mousePos.y = HIWORD(lParam);
//		ClientToScreen(hwnd, &mousePos);
		return 0;

	case WM_SIZE:
		resize(LOWORD(lParam), HIWORD(lParam));
		return 0;

	case WM_SYSCOMMAND:
		switch(wParam)
		{
		case SC_SCREENSAVE:
		case SC_MONITORPOWER:
			return 0;
		}
		break;

	case WM_DESTROY:
		if (hglrc)
		{
			glDeleteTextures(2, texObject);
			glDeleteLists(1000, 256);
			wglMakeCurrent(NULL, NULL);
			wglDeleteContext(hglrc);
			ReleaseDC(hwnd, hdc);
		}
		gPhysicsSDK->release();
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}

void resize(GLint width, GLint height)
{
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (!height)
		height = 1;
	gluPerspective( 45.0, (double) width / height, 1.0, 1000.0 );
}

BOOL setupPixelFormat(HDC hdc)
{
	PIXELFORMATDESCRIPTOR pfd;
	INT pixelformat;

	pfd.nSize		= sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion		= 1;
	pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_GENERIC_ACCELERATED | PFD_DOUBLEBUFFER;
	pfd.dwLayerMask		= PFD_MAIN_PLANE;
	pfd.iPixelType		= PFD_TYPE_RGBA;
	pfd.cColorBits		= 32;
	pfd.cDepthBits		= 24;
	pfd.cAccumBits		= 32;
	pfd.cStencilBits	= 32;

	if ( (pixelformat = ChoosePixelFormat(hdc, &pfd)) == 0 )
	{
		MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
		return FALSE;
	}

	if (SetPixelFormat(hdc, pixelformat, &pfd) == FALSE)
	{
		MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
		return FALSE;
	}
	return TRUE; 
}

BOOL setupNewPixelFormat(HDC hdc)
{
	PIXELFORMATDESCRIPTOR	pfd;
	FLOAT			fList[] = {0, 0};
	UINT			numFormats;
	INT			nPixelFormat,
				iList[] = {	WGL_DOUBLE_BUFFER_ARB, TRUE, WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
						WGL_SAMPLE_BUFFERS_ARB, 1, WGL_SAMPLES_ARB, 4, 0, 0 },
				iResult[3];

//BOOL  wglChoosePixelFormatARB(HDC hdc, const int *iList, const FLOAT *fList, UINT maxReturned, int *pixelformathandle, UINT *numMatching);
	wglChoosePixelFormatARB(hdc, iList, fList, 1, &nPixelFormat, &numFormats);

	iList[1] = WGL_ACCELERATION_ARB;
	iList[2] = 0;
//BOOL  wglGetPixelFormatAttribivARB(HDC hdc, int iPixelFormat, int iLayerPlane, UINT nAttributes, const int *piAttributes, FLOAT *pfValues);
//	wglGetPixelFormatAttribivARB(hdc, nPixelFormat, 0, 2, iList, iResult);

	SetPixelFormat(hdc, nPixelFormat, &pfd);
	return TRUE;
}

int initExt(void)
{
	const GLubyte	*sVer;

	sVer = glGetString(GL_VERSION);

	if (sVer[0] < 2)
		return 1;

	glGenProgramsARB	= (PFNGLGENPROGRAMSARBPROC) wglGetProcAddress("glGenProgramsARB");
	glBindProgramARB	= (PFNGLBINDPROGRAMARBPROC) wglGetProcAddress("glBindProgramARB");
	glProgramStringARB	= (PFNGLPROGRAMSTRINGARBPROC) wglGetProcAddress("glProgramStringARB");
	glActiveTexture		= (PFNGLACTIVETEXTUREPROC) wglGetProcAddress("glActiveTexture");

	//GLSL
	glShaderSource		= (PFNGLSHADERSOURCEPROC) wglGetProcAddress("glShaderSource");
	glCompileShader		= (PFNGLCOMPILESHADERPROC) wglGetProcAddress("glCompileShader");
	glCreateShader		= (PFNGLCREATESHADERPROC) wglGetProcAddress("glCreateShader");
	glGetProgramInfoLog	= (PFNGLGETPROGRAMINFOLOGPROC) wglGetProcAddress("glGetProgramInfoLog");
	glGetShaderInfoLog	= (PFNGLGETSHADERINFOLOGPROC) wglGetProcAddress("glGetShaderInfoLog");
	glCreateProgram		= (PFNGLCREATEPROGRAMPROC) wglGetProcAddress("glCreateProgram");
	glAttachShader		= (PFNGLATTACHSHADERPROC) wglGetProcAddress("glAttachShader");
	glLinkProgram		= (PFNGLLINKPROGRAMPROC) wglGetProcAddress("glLinkProgram");
	glValidateProgram	= (PFNGLVALIDATEPROGRAMPROC) wglGetProcAddress("glValidateProgram");
	glUseProgram		= (PFNGLUSEPROGRAMPROC) wglGetProcAddress("glUseProgram");
	glDetachShader		= (PFNGLDETACHSHADERPROC) wglGetProcAddress("glDetachShader");
	glDeleteShader		= (PFNGLDELETESHADERPROC) wglGetProcAddress("glDeleteShader");
	glGetShaderiv		= (PFNGLGETSHADERIVPROC) wglGetProcAddress("glGetShaderiv");
	glGetProgramiv		= (PFNGLGETPROGRAMIVPROC) wglGetProcAddress("glGetProgramiv");
	glGetUniformLocation	= (PFNGLGETUNIFORMLOCATIONPROC) wglGetProcAddress("glGetUniformLocation");
	glUniform1i		= (PFNGLUNIFORM1IPROC) wglGetProcAddress("glUniform1i");
	glUniform1f		= (PFNGLUNIFORM1FPROC) wglGetProcAddress("glUniform1f");
	glUniform2f		= (PFNGLUNIFORM2FPROC) wglGetProcAddress("glUniform2f");
	glUniform3f		= (PFNGLUNIFORM3FPROC) wglGetProcAddress("glUniform3f");
	glUniform4f		= (PFNGLUNIFORM4FPROC) wglGetProcAddress("glUniform4f");

PFNGLGETUNIFORMLOCATIONPROC		glGetUniformLocation;
PFNGLUNIFORM1FPROC			glUniform1f;


//	function		= (cast) wglGetProcAddress("function");
	return 0;
}

CHAR *getShader(CHAR *fileName)
{
	HANDLE		hFile;
	LARGE_INTEGER	fSize;
	CHAR		*file;
	ULONG		numRead;

	hFile = CreateFile(fileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) 
		return 0;

	if(!GetFileSizeEx(hFile, &fSize))
		return 0;

	file = (char *)malloc(sizeof(char) * fSize.LowPart);

	file[fSize.LowPart] = '\0';

	if (!ReadFile(hFile, file, fSize.LowPart, &numRead, NULL))
		return 0;

	CloseHandle(hFile);
	if (fSize.LowPart != numRead)
		return 0;

	return file;
}